import type Clutter from 'gi://Clutter';

import { registerClass } from "./utils.js";

type Constructor<T = {}> = new (...args: any[]) => T;
// Mixins need to cache the created classes or else they would register the same class name
// multiple times, which isn't supported by GObject
interface CachingMixin<T> {
	cache?: {
		[index: string]: Constructor<T>
	}
}

export const AutoHidable = <CachingMixin<Clutter.Actor>>function <TBase extends Constructor<Clutter.Actor>>(Base: TBase) {
	if (AutoHidable.cache === undefined) AutoHidable.cache = {};
	if (AutoHidable.cache[Base.name] !== undefined) return AutoHidable.cache[Base.name];

	const cls = registerClass({
		GTypeName: `AutoHidable_${Base.name}`,
	}, class extends Base {
		#_container?: Clutter.Actor;

		constructor(...args: any[]) {
			const container = args.at(-1).container as Clutter.Actor | null | undefined;
			delete args.at(-1).container;
			super(...args);

			// if `container` is null, don't do anything
			if (container === undefined) {
				this.container = this;
			} else if (container !== null) {
				this.container = container;
			}
		}

		get container(): Clutter.Actor | undefined {
			return this.#_container;
		}

		set container(value: Clutter.Actor | undefined) {
			if (this.#_container !== undefined) this.disconnect_named(this.#_container);
			if (value !== null) {
				this.#_container = value;
				this.connect_named(this.#_container, 'child-added', (_container, child) => {
					this.connect_named(child, 'notify::visible', this._update_visibility.bind(this));
					this._update_visibility();
				});
				this.connect_named(this.#_container, 'child-removed', (_container, child) => {
					this.disconnect_named(child);
					this._update_visibility();
				});
				this._update_visibility();
			} else {
				this.#_container = undefined;
			}
		}

		_get_ah_children() {
			return this.#_container?.get_children();
		}

		_update_visibility() {
			for (const child of this._get_ah_children()) {
				if (child.visible) {
					this.show();
					return;
				}
			}

			this.hide();
			// Force the widget to take no space when hidden (this fixes some bugs but I don't know why)
			this.queue_relayout();
		}
	});

	AutoHidable.cache[Base.name] = cls;
	return cls;
};
